class Version:
    def __init__(self, major, minor, patch):
        self.__major = major
        self.__minor = minor
        self.__patch = patch

    @classmethod
    def parse(cls, string):
        try:
            nums = [int(num) for num in string.split(".")]
        except ValueError:
            raise click.Exception("invalid version format")

        if len(nums) == 1:
            return cls(nums[0], 0, 0)

        elif len(nums) == 2:
            return cls(nums[0], nums[1], 0)

        elif len(nums) == 3:
            return cls(nums[0], nums[1], nums[2])

        else:
            raise click.Exception("invalid version format")

    @property
    def major(self):
        return self.__major

    @property
    def minor(self):
        return self.__minor

    @property
    def patch(self):
        return self.__patch

    def __eq__(self, other):
        if not isinstance(other, Version):
            return NotImplemented

        return (
            self.__major == other.__major
            and self.__minor == other.__minor
            and self.__patch == other.__patch
        )

    def __lt__(self, other):
        if not isinstance(other, Version):
            return NotImplemented

        if self.__major == other.__major:
            if self.__minor == other.__minor:
                return self.__patch < other.__patch
            else:
                return self.__minor < other.__minor
        else:
            return self.__major < other.__major

    def __le__(self, other):
        if not isinstance(other, Version):
            return NotImplemented

        if self.__major == other.__major:
            if self.__minor == other.__minor:
                return self.__patch <= other.__patch
            else:
                return self.__minor < other.__minor
        else:
            return self.__major < other.__major

    def __repr__(self):
        return f"{type(self).__module__}.{type(self).__qualname__}({self.major}, {self.minor}, {self.patch})"

    def __str__(self):
        if self.patch:
            return f"{self.major}.{self.minor}.{self.patch}"
        else:
            if self.minor:
                return f"{self.major}.{self.minor}"
            else:
                return f"{self.major}"



import click
import importlib
import os, shutil
import jinja2
import json
from pathlib import Path

from . import log
from .plugin import PluginScheduler, LogProxy
from .snapshot import Snapshot, Diff
from .markdown import Markdown
from .utils import file_status, file_status_done, cleartree


class Builder:
    def __init__(self, config, base_dir, **options):
        self.__config = config
        self.__options = options

        self.__base_dir = base_dir

        self.__snapshots = {
            "source": {"path": self.source_dir},
            "templates": {"path": self.templates_dir},
            "static": {"path": self.static_dir},
        }
        self.__templates = {}

        self.__md = Markdown()
        self.__j2 = jinja2.Environment(
            loader=jinja2.FileSystemLoader(str(self.templates_dir))
        )

        self.__plugin_packages = {}

    @property
    def cache_dir(self):
        return self.__base_dir / ".cache"

    @property
    def output_dir(self):
        return self.__base_dir / self.__config.output_directory

    @property
    def static_output_dir(self):
        return self.output_dir / "_static"

    @property
    def source_dir(self):
        return self.__base_dir / self.__config.source_directory

    @property
    def templates_dir(self):
        return self.__base_dir / self.__config.templates_directory

    @property
    def static_dir(self):
        return self.__base_dir / self.__config.static_directory

    def build(self):
        PluginScheduler.set_context(self)

        self.__load_plugins()

        PluginScheduler.set_config(
            {
                plugin: self.__config.plugins[plugin].get("config", {})
                for plugin in self.__config.plugins
            }
        )
        PluginScheduler.setup()

        self.__scan_directories()

        if self.__options["fresh"]:
            self.__clear_output_directory()
        else:
            self.__load_cache_data()

        PluginScheduler.build_started()
        click.echo("Build started ...")

        self.__render_pages()
        self.__copy_static_files()

        PluginScheduler.build_ended()

        self.__dump_cache_data()

    def snapshot_register(self, name, path):
        self.__snapshots[name] = {"path": self.__base_dir / path}

    def snapshot_current(self, name):
        return self.__snapshots[name]["current"]

    def snapshot_old(self, name):
        return self.__snapshots[name].get("old", Snapshot({}))

    def snapshot_diff(self, name):
        return self.snapshot_current(name).diff(self.snapshot_old(name))

    def get_package_alias(self, pkg, default=None):
        return self.__plugin_packages.get(pkg, default)

    def __load_plugins(self):
        for name, plugin in self.__config.plugins.items():
            if "package" in plugin:
                # loading installed package
                self.__plugin_packages[plugin["package"]] = name
                try:
                    importlib.import_module(plugin["package"])
                except Exception as e:
                    log.error(f"can't load plugin “{name}”: {e}")
                    raise click.ClickException("failed to load plugins")

            elif "script" in plugin:
                # loading module from file path
                script_path = Path(plugin["script"])
                if not script_path.is_absolute():
                    script_path = self.__base_dir / script_path

                spec = importlib.util.spec_from_file_location(
                    name + "_komoe_plugin", script_path
                )
                module = importlib.util.module_from_spec(spec)

                try:
                    spec.loader.exec_module(module)
                except Exception as e:
                    log.error(f"can't load plugin “{name}”: {e}")
                    raise click.ClickException("failed to load plugins")

            else:
                log.warn(f"plugin “{name}” is declared but has no package/script")

    def __load_cache_data(self):
        # loading previous snapshots
        for name in self.__snapshots:
            snapshot_path = self.cache_dir / ("snapshot_" + name)
            if snapshot_path.is_file():
                with open(snapshot_path, "rt", encoding="utf8") as f:
                    self.__snapshots[name]["old"] = Snapshot.load(f.read())

        # load previous page-template relationships
        relationships_path = self.cache_dir / "relationships"
        if relationships_path.is_file():
            with open(relationships_path, "rt", encoding="utf8") as f:
                self.__templates = json.load(f)

    def __dump_cache_data(self):
        if not self.cache_dir.exists():
            self.cache_dir.mkdir()

        # dump snapshots
        for name in self.__snapshots:
            snapshot_path = self.cache_dir / ("snapshot_" + name)
            with open(snapshot_path, "wt+", encoding="utf8") as f:
                f.write(self.__snapshots[name]["current"].dump())

        # dump page-template relationships
        with open(self.cache_dir / "relationships", "wt+", encoding="utf8") as f:
            json.dump(self.__templates, f)

    def __scan_directories(self):
        for name in self.__snapshots:
            self.__snapshots[name]["current"] = Snapshot.scan(
                self.__snapshots[name]["path"]
            )

    def __render_pages(self):
        created = list()
        modified = list()
        removed = list()
        same = list()

        for file, diff in self.snapshot_diff("source").items():
            if diff == Diff.CREATED:
                created.append(file)
            elif diff == Diff.DELETED:
                removed.append(file)
            elif diff == Diff.MODIFIED:
                modified.append(file)
            elif diff == Diff.SAME:
                same.append(file)

        if same:
            need_refresh = list()
            for file, diff in self.snapshot_diff("templates").items():
                if diff == Diff.MODIFIED:
                    need_refresh += self.__templates.get(file, [])
            for file in same:
                if file in need_refresh:
                    modified.append(file)

        env_info = ", ".join(
            ([f"{len(created)} added"] if created else [])
            + ([f"{len(modified)} modified"] if modified else [])
            + ([f"{len(removed)} removed"] if removed else [])
        )
        if env_info:
            log.info(f"Document environment updated: {env_info}")
        else:
            log.info(f"Document environment updated: no changes")
            return

        click.echo("Rendering pages:")

        for file in created:
            self.__render_page(file, Diff.CREATED)

        for file in modified:
            self.__render_page(file, Diff.MODIFIED)

        for file in removed:
            self.__remove_page(file)

    def __page_location(self, file):
        base, _ = os.path.splitext(file)
        dest = base + ".html"

        return (
            self.source_dir / file,
            self.output_dir / dest,
        )

    def __render_page(self, file, modified):
        file_status(file, modified)

        src_path, dst_path = self.__page_location(file)

        depth = len(Path(file).parts) - 1
        rel_root = "/".join([".."] * depth) if depth else "."

        with open(src_path, "rt", encoding="utf8") as f:
            md = f.read()

        content = self.__md.render(md)

        template_path = str(self.__find_template_file().relative_to(self.templates_dir))

        same_template = False
        for old_template, sources in self.__templates.items():
            if file in sources:
                if template_path == old_template:
                    same_template = True
                else:
                    sources.remove(file)
                break
        if not same_template:
            if template_path in self.__templates:
                self.__templates[template_path].append(str(file))
            else:
                self.__templates[template_path] = [str(file)]

        tpl = self.__j2.get_template(template_path)

        title = self.__md.document_title
        if "title" in self.__md.metadata:
            title = self.__md.metadata["title"][0]

        shared = {
            "TITLE": title,
            "ROOT": rel_root,
            "STATIC": rel_root + "/_static",
        }

        content_tpl = self.__j2.from_string(content)
        html = tpl.render(CONTENT=content_tpl.render(**shared), **shared)

        os.makedirs(dst_path.parent, exist_ok=True)

        with open(dst_path, "wt+", encoding="utf8") as f:
            f.write(html)

        file_status_done()

    def __remove_page(self, file):
        file_status(file, Diff.REMOVED)

        _, path = self.__page_location(file)

        self.__templates_remove_source(file)

        try:
            os.remove(self.output_dir / path)
            click.echo("done")
        except FileNotFoundError:
            print()
            log.warn(f"Failed to remove {path}: file not found")

        file_status_done()

    def __copy_static_files(self):
        created = list()
        modified = list()
        removed = list()

        for file, diff in self.snapshot_diff("static").items():
            if diff == Diff.CREATED:
                created.append(file)
            elif diff == Diff.DELETED:
                removed.append(file)
            elif diff == Diff.MODIFIED:
                modified.append(file)

        env_info = ", ".join(
            ([f"{len(created)} added"] if created else [])
            + ([f"{len(modified)} modified"] if modified else [])
            + ([f"{len(removed)} removed"] if removed else [])
        )
        if env_info:
            log.info(f"Static environment updated: {env_info}")
        else:
            log.info(f"Static environment updated: no changes")
            return

        click.echo("Copying static files:")

        for file in created:
            self.__copy_static_file(file, Diff.CREATED)

        for file in modified:
            self.__copy_static_file(file, Diff.MODIFIED)

        for file in removed:
            self.__remove_static_file(file)

    def __copy_static_file(self, file, modified):
        file_status(file, modified)

        dest = self.static_output_dir / file

        os.makedirs(dest.parent, exist_ok=True)
        shutil.copy(self.static_dir / file, dest)

        file_status_done()

    def __remove_static_file(self, file):
        file_status(file, Diff.DELETED)

        try:
            os.remove(self.static_output_dir / file)
            file_status_done()
        except FileNotFoundError:
            print()
            log.warn(f"Failed to remove {file}: file not found")

    def __find_template_file(self):
        directory, name = os.path.split(self.__md.template)
        directory = self.templates_dir / directory

        if not directory.is_dir():
            log.error(f"No such template : {self.__md.template}")
            raise click.ClickException(f"failed to render {file}")

        path = None
        for f in directory.iterdir():
            if f.is_file() and f.name.startswith(name + "."):
                path = f
        if path is None:
            log.error(f"No such template : {self.__md.template}")
            raise click.ClickException(f"failed to render {file}")

        return path

    def __clear_output_directory(self):
        if self.output_dir.is_dir():
            cleartree(self.output_dir)



from pathlib import Path
from enum import Enum, auto


class Diff(Enum):
    CREATED = auto()
    MODIFIED = auto()
    SAME = auto()
    DELETED = auto()


def _scan(root, path, ignore_hidden, ignore_patterns):
    files = dict()

    for e in path.iterdir():
        if ignore_hidden and e.name.startswith("."):
            continue

        ignore = False
        for pattern in ignore_patterns:
            if e.match(pattern):
                ignore = True
        if ignore:
            continue

        if e.is_file():
            files[str(e.relative_to(root))] = int(e.stat().st_mtime)

        elif e.is_dir():
            files.update(_scan(root, e, ignore_hidden, ignore_patterns))

    return files


class Snapshot:
    def __init__(self, files):
        self.__files = files

    @classmethod
    def scan(cls, root, ignore_hidden=True, ignore_patterns=[]):
        if not isinstance(root, Path):
            root = Path(root)

        if not root.is_dir():
            raise ValueError("root must be an existing directory")

        return cls(_scan(root, root, ignore_hidden, ignore_patterns))

    @classmethod
    def load(cls, text):
        data = dict()
        for entry in text.split("\n"):
            if len(entry) == 0:
                continue
            path, time = entry.rsplit(":", 1)
            time = int(time)
            data[path] = time
        return cls(data)

    def dump(self):
        text = str()
        for path, time in self.__files.items():
            text += f"{path}:{time}\n"
        return text

    def diff(self, old):
        diff_dict = {}

        for entry in set(self.__files) | set(old.__files):
            if entry in self.__files:
                if entry in old.__files:
                    if self.__files[entry] == old.__files[entry]:
                        diff_dict[entry] = Diff.SAME
                    else:
                        diff_dict[entry] = Diff.MODIFIED
                else:
                    diff_dict[entry] = Diff.CREATED
            else:
                diff_dict[entry] = Diff.DELETED

        return diff_dict



import os

CONFIG = """\
# Configuration for KOMOE static site generator.

komoe_require = '0.2.2'

[project]
name = '{0}'

[build]
source = 'source'
templates = 'templates'
static = 'static'
output = 'build'
"""

BASE = """\
<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{{ TITLE }}</title>
</head>
<body>
    {{ CONTENT }}
</body>
</html>
"""

PAGE = """\
@base

# {0}

Welcome to the Komoe quickstart !
"""


def create_new_project(path, name):
    with open(path / "komoe.toml", "wt+", encoding="utf8") as f:
        f.write(CONFIG.format(name))

    os.mkdir(path / "source")

    with open(path / "source" / "index.md", "wt+", encoding="utf8") as f:
        f.write(PAGE.format(name))

    os.mkdir(path / "templates")

    with open(path / "templates" / "base.j2.html", "wt+", encoding="utf8") as f:
        f.write(BASE)

    os.mkdir(path / "static")



import click
import tomli

from . import log
from .version import Version


def _require(cfg, *path):
    for key in path:
        if key in cfg:
            cfg = cfg[key]
        else:
            log.error(f"{'.'.join(path)} is missing from configuration file")
            raise click.ClickException("invalid configuration file")
    return cfg


def _default(cfg, default, *path):
    for key in path:
        if key in cfg:
            cfg = cfg[key]
        else:
            return default
    return cfg


class ProjectConfig:
    def __init__(self, cfg):
        self.__minimum_required_version = Version.parse(_require(cfg, "komoe_require"))

        self.__source_dir = _require(cfg, "build", "source")
        self.__templates_dir = _require(cfg, "build", "templates")
        self.__static_dir = _require(cfg, "build", "static")
        self.__output_dir = _require(cfg, "build", "output")

        self.__project_infos = _default(cfg, {}, "project")
        self.__plugins = _default(cfg, {}, "plugin")

    @classmethod
    def from_file(cls, path):
        try:
            with open(path, "rb") as f:
                toml_dict = tomli.load(f)
        except tomli.TOMLDecodeError as e:
            log.error(f"error reading configuration file: {e}")
            raise click.ClickException("invalid configuration file")

        return cls(toml_dict)

    @property
    def minimum_required_version(self):
        return self.__minimum_required_version

    @property
    def source_directory(self):
        return self.__source_dir

    @property
    def templates_directory(self):
        return self.__templates_dir

    @property
    def static_directory(self):
        return self.__static_dir

    @property
    def output_directory(self):
        return self.__output_dir

    @property
    def project(self):
        return self.__project_infos

    @property
    def plugins(self):
        return self.__plugins



import os, shutil, pathlib
import click

from .snapshot import Diff


def file_status(filename, diff):
    tag = "   "
    if diff == Diff.CREATED:
        tag = " + "
    elif diff == Diff.MODIFIED:
        tag = " ~ "
    elif diff == Diff.SAME:
        tag = " = "
    elif diff == Diff.DELETED:
        tag = " - "

    click.secho(tag, nl=False, bold=True)
    click.echo(f"{filename} … ", nl=False)


def file_status_done():
    click.echo("\x1b[2D✓")


def cleartree(path):
    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)

    for item in path.iterdir():
        if item.is_file():
            os.remove(item)
        elif item.is_dir():
            shutil.rmtree(item)



import click

from . import log
from .snapshot import Diff
from .utils import file_status, file_status_done

__all__ = [
    "before_build",
    "after_build",
    "before_plugin",
    "after_plugin",
    "setup",
    "Diff",
    "file_status",
    "file_status_done",
]


class _ModuleEvents:
    def __init__(self):
        self.__on_start = list()
        self.__on_end = list()

    def register(self, event, action):
        if event == "start":
            self.__on_start.append(action)
        elif event == "end":
            self.__on_end.append(action)
        else:
            raise ValueError(f"invalid event “{event}”")

    def on(self, event):
        if event == "start":
            return iter(self.__on_start)
        elif event == "end":
            return iter(self.__on_end)
        else:
            raise ValueError(f"invalid event “{event}”")


class _Action:
    def __init__(self, callback):
        self.__callback = callback
        self.__called = False
        self.__module = None

    def __bool__(self):
        return self.__called

    def __call__(self, context, config):
        if self.__called:
            raise RuntimeError("Action object called twice")
        else:
            self.__called = True
            return self.__callback(context, config)

    @property
    def module(self):
        return self.__module

    @module.setter
    def module(self, value):
        self.__module = value


class _ModuleActions:
    def __init__(self, name):
        self.__actions = list()
        self.__name = name

    def add(self, action):
        action.module = self
        self.__actions.append(action)

    @property
    def started(self):
        return any(self.__actions)

    @property
    def ended(self):
        return all(self.__actions)

    @property
    def name(self):
        return self.__name


class PluginScheduler:
    __setup = []
    __events = {"build!": _ModuleEvents()}
    __actions = {}

    __context = None
    __config = None

    @classmethod
    def events(cls, module):
        if not module in cls.__events:
            cls.__events[module] = _ModuleEvents()
        return cls.__events[module]

    @classmethod
    def actions(cls, module):
        if not module in cls.__actions:
            cls.__actions[module] = _ModuleActions(module)

        return cls.__actions[module]

    @classmethod
    def subscribe(cls, module, event, callback):
        plugin_name = cls.__context.get_package_alias(
            callback.__module__,
            callback.__module__.replace("_komoe_plugin", ""),
        )
        action_name = callback.__name__

        if plugin_name == module:
            log.error(
                f"a plugin can't subscribe to itself ({plugin_name}.{callback.__name__})"
            )
            raise click.ClickException("failed to load plugins")

        action = _Action(callback)

        events = cls.events(module)
        events.register(event, action)

        actions = cls.actions(plugin_name)
        actions.add(action)

        cls.events(plugin_name)

    @classmethod
    def register_setup(cls, callback):
        plugin_name = cls.__context.get_package_alias(
            callback.__module__,
            callback.__module__.replace("_komoe_plugin", ""),
        )

        cls.__setup.append((plugin_name, callback))

    @classmethod
    def build_started(cls):
        cls.notify("build!", "start")

    @classmethod
    def build_ended(cls):
        cls.notify("build!", "end")

    @classmethod
    def set_context(cls, context):
        cls.__context = context

    @classmethod
    def set_config(cls, config):
        cls.__config = config

    @classmethod
    def notify(cls, module, event):
        for action in cls.__events[module].on(event):
            if not action.module.started:
                cls.notify(action.module.name, "start")

            action(
                BuilderProxy(cls.__context, action.module.name),
                cls.__config.get(action.module.name, {}),
            )

            if action.module.ended:
                cls.notify(action.module.name, "end")

    @classmethod
    def setup(cls):
        for module, callback in cls.__setup:
            callback(
                BuilderProxy(cls.__context, module),
                cls.__config.get(module, {}),
            )


class LogProxy:
    def __init__(self, ctx):
        self.__context = ctx

    @property
    def context(self):
        return self.__context

    def error(self, message):
        log.error(f"{self.context}: {message}")

    def warn(self, message):
        log.warn(f"{self.context}: {message}")

    def info(self, message):
        log.info(f"{self.context}: {message}")


class BuilderProxy:
    def __init__(self, builder, log_ctx):
        self.__builder = builder
        self.__log = LogProxy(log_ctx)

    @property
    def log(self):
        return self.__log

    @property
    def echo(self):
        return click.echo

    def fatal(self, message=None):
        raise click.ClickException(
            f"plugin {self.__log.context} aborted the build"
            + ("" if message is None else ": " + message)
        )

    def snapshot_register(self, name, path):
        return self.__builder.snapshot_register(name, path)

    def snapshot_current(self, name):
        return self.__builder.snapshot_current(name)

    def snapshot_old(self, name):
        return self.__builder.snapshot_old(name)

    def snapshot_diff(self, name):
        return self.__builder.snapshot_diff(name)

    @property
    def cache_dir(self):
        return self.__builder.cache_dir

    @property
    def output_dir(self):
        return self.__builder.output_dir

    @property
    def static_output_dir(self):
        return self.__builder.static_output_dir

    @property
    def source_dir(self):
        return self.__builder.source_dir

    @property
    def templates_dir(self):
        return self.__builder.templates_dir

    @property
    def static_dir(self):
        return self.__builder.static_dir


def setup(func):
    PluginScheduler.register_setup(func)


def before_build(func):
    PluginScheduler.subscribe("build!", "start", func)


def after_build(func):
    PluginScheduler.subscribe("build!", "end", func)


def before_plugin(plugin_name):
    def deco(func):
        PluginScheduler.subscribe(plugin_name, "start", func)

    return deco


def after_plugin(plugin_name):
    def deco(func):
        PluginScheduler.subscribe(plugin_name, "end", func)

    return deco



import click


def error(message):
    __log("ERROR", "red", message)


def warn(message):
    __log("WARNING", "yellow", message)


def info(message):
    __log("INFO", None, message)


def __log(level, color, message):
    click.secho(f"[{level}] ", nl=False, err=True, fg=color, bold=True)
    click.secho(message, err=True, fg=color)



from setuptools import setup

with open("README.md", "rt", encoding="utf8") as f:
    README = f.read()

setup(
    name="Komoe",
    version="0.2.2",
    description="A static site generator",
    long_description=README,
    long_description_content_type="text/markdown",
    license="MIT License",
    author="Louis DEVIE",
    url="https://github.com/louisdevie/komoe",
    packages=["komoe"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=[
        "static",
        "site",
        "website",
        "template",
        "templating",
    ],
    python_requires=">=3.8",
    install_requires=[
        "Jinja2>=3.1.2",
        "Markdown==3.4.1",
        "click>=8.1.3",
        "tomli>=2.0.1",
    ],
    entry_points={
        "console_scripts": [
            "komoe = komoe.commands:main",
        ]
    },
)



import click
import os
from pathlib import Path

from . import template, __version__
from .config import ProjectConfig
from .builder import Builder


@click.group()
def main():
    pass


@main.command()
@click.argument(
    "path",
    default=".",
    type=click.Path(file_okay=False, path_type=Path),
)
@click.option(
    "--project-name", "-N", prompt=True, required=True, help="The name of the project"
)
def new(path, project_name):
    """Creates a new project

    If PATH isn't specified, the current directory is used.
    """

    if path.exists():
        entries = [
            entry for entry in path.iterdir() if not entry.name.startswith(".git")
        ]
        if len(entries) != 0:
            path_repr = str(path)
            if path_repr == ".":
                path_repr = "the current directory"
            raise click.ClickException(f"{path_repr} isn't empty")
    else:
        os.makedirs(path)

    template.create_new_project(path, project_name)


@main.command()
@click.option(
    "--project-file",
    "-p",
    type=click.Path(dir_okay=False, exists=True, path_type=Path),
    help="Build a specific project file (overrides --project-dir)",
)
@click.option(
    "--project-dir",
    "-P",
    type=click.Path(file_okay=False, exists=True, path_type=Path),
    help="Build the project in that directory",
)
@click.option("--fresh", is_flag=True, help="Regenerates all content")
def build(project_file, project_dir, fresh):
    """Build a project

    If no project is specified, the project in the current directory will be built.
    """

    if project_file is not None:
        config_path = project_file

    elif (project_dir is not None) and (
        "komoe.toml" in f.name for f in project_dir.iterdir()
    ):
        config_path = project_dir / "komoe.toml"

    elif "komoe.toml" in (f.name for f in Path.cwd().iterdir()):
        config_path = Path.cwd() / "komoe.toml"

    else:
        raise click.ClickException("project file not found")

    config = ProjectConfig.from_file(config_path)

    if config.minimum_required_version > __version__:
        raise click.ClickException(
            f"The project requires at least Komoe v{config.minimum_required_version}"
        )

    builder = Builder(config, config_path.parent, fresh=fresh)

    builder.build()



import markdown


class Markdown:
    def __init__(self):
        self.__md = markdown.Markdown(
            extensions=[
                MarkdownExtension(),
                "attr_list",
                "fenced_code",
                "footnotes",
                "tables",
                "admonition",
                "meta",
                "sane_lists",
                "smarty",
                "toc",
            ]
        )
        self.__md.komoe = self

        self.__template = None
        self.__title = ""

    def render(self, text):
        self.__md.reset()
        return self.__md.convert(text)

    @property
    def template(self):
        return self.__template

    @template.setter
    def template(self, value):
        self.__template = value

    @property
    def document_title(self):
        return self.__title

    @document_title.setter
    def document_title(self, value):
        self.__title = value

    @property
    def metadata(self):
        return self.__md.Meta


class MarkdownExtension(markdown.Extension):
    def extendMarkdown(self, md):
        self.__md = md
        self.__md.registerExtension(self)

        self.__md.preprocessors.register(
            TemplatePreprocessor(self.__md),
            "komoe.preprocessor.template",
            200,
        )
        self.__md.treeprocessors.register(
            TitleTreeprocessor(self.__md), "komoe.treeprocessor.title", 200
        )

    def reset(self):
        self.__md.template = None
        self.__md.document_title = ""


class TemplatePreprocessor(markdown.preprocessors.Preprocessor):
    def run(self, lines):
        new_lines = lines.copy()

        while not new_lines[0].strip():
            new_lines.pop(0)

        if new_lines[0].startswith("@"):
            self.md.komoe.template = new_lines[0][1:]
            new_lines.pop(0)

        return new_lines

    def reset(self):
        self.md.komoe.template = None


class TitleTreeprocessor(markdown.treeprocessors.Treeprocessor):
    def run(self, root):
        self.md.komoe.document_title = root.find("h1").text
